# coding: utf-8
"""
Content
-----------------------------------------------------------------------------
This module exposes basic API of mzml2isa, either being called from command
line interface with arguments parsing via **main** function, or from another
Python program via the **fparse** function which works the same.

About
-----------------------------------------------------------------------------
The mzml2isa parser was created by Tom Lawson (University of Birmingham, UK)
as part of a NERC funded placement at EBI Cambridge in June 2015. Python 3
port and enhancements were carried out by Martin Larralde (ENS Cachan, FR)
in June 2016 during an internship at the EBI Cambridge.

License
-----------------------------------------------------------------------------
GNU General Public License version 3.0 (GPLv3)
"""
from __future__ import absolute_import

import io
import os
import sys
import six
import glob
import argparse
import textwrap
import warnings
import json
import tarfile
import zipfile
import multiprocessing
import multiprocessing.pool
import pronto
import functools

try:
    import progressbar
except ImportError:
    progressbar = None

from . import (
    __author__,
    __version__,
    __name__,
    __license__,
)
from .isa   import ISA_Tab
from .mzml  import mzMLmeta, imzMLmeta
from .usermeta import UserMetaLoader
from .utils import (
    longest_substring,
    merge_spectra,
    compr_extract,
    star_args,
    get_ontology,
)


@star_args
def _parse_file(filepath, ontology, parser, pbar=None):
    """Parse a single file using a cache ontology and a metadata extractor

    Arguments:
        filepath (str): path to the mzml/imzml file to parse
        ontology (pronto.Ontology): the cached ontology to use
            (either IMS or MS)
        parser (mzml.mzMLmeta): the parser to use on the file
            (either mzml2isa.mzml.mzMLmeta or mzml2isa.mzml.imzMLmeta)
        pbar (progressbar.ProgressBar, optional): a progressbar
            to display progresses onto [default: None]

    Returns:
        dict: a dictionary containing the extracted metadata
    """
    meta = parser(filepath, ontology).meta
    if pbar is not None:
        pbar.update(pbar.value + 1)
    else:
        print("Finished parsing: {}".format(filepath))
    return meta

def convert(in_path, out_path, study_identifier, **kwargs):
    """ Parses a study from given *in_path* and then creates an ISA file.

    A new folder is created in the out directory bearing the name of
    the study identifier.

    Arguments:
        in_path (str): path to the directory or archive containing mzml files
        out_path (str): path to the output directory (new directories will be
            created here)
        study_identifier (str): study identifier (e.g. MTBLSxxx)

    Keyword Arguments:
        usermeta (str, optional): the path to a json file, a xlsx file or
            directly a json formatted string containing user-defined
            metadata [default: None]
        split (bool, optional): split assay files based on the polarity
            of the scans [default: True]
        merge (bool, optional): for imzML studies, try to merge centroid
            and profile scans in a single sample row [default: False]
        jobs (int, optional): the number of jobs to use for parsing
            mzML or imzML files [default: 1]
        template_directory (str, optional): the path to a directory
            containing custom templates to use when importing ISA tab
            [default: None]
        verbose (bool): display more output [default: True]
    """
    split = kwargs.get('split', True)
    merge = kwargs.get('merge', False)
    verbose = kwargs.get('verbose', True)
    jobs = kwargs.get('jobs', 1)
    template_directory = kwargs.get('template_directory', None)

    PARSERS = {'mzML': mzMLmeta, 'imzML': imzMLmeta}
    ONTOLOGIES = {'mzML': get_ontology('MS'), 'imzML': get_ontology('IMS')}

    # open user metadata file if any
    meta_loader = UserMetaLoader(kwargs.get('usermeta', None))

    # get mzML file in the example_files folder
    if os.path.isdir(in_path):
        compr = False
        mzml_files = glob.glob(os.path.join(in_path, "*mzML"))
    elif tarfile.is_tarfile(in_path) or zipfile.is_zipfile(in_path):
        compr = True
        mzml_files = compr_extract(in_path)
    else:
        raise SystemError("Couldn't recognise format of "
                          "{} as a source of mzml files".format(in_dir))

    if mzml_files:
        # store the first mzml_files extension
        extension = getattr(mzml_files[0], 'name', mzml_files[0]).split(os.path.extsep)[-1]
        ontology = ONTOLOGIES[extension]
        parser = PARSERS[extension]

        if not verbose and progressbar is not None:
            pbar = progressbar.ProgressBar(
                min_value = 0, max_value = len(mzml_files),
                widgets=['Parsing {:8}: '.format(study_identifier),
                           progressbar.SimpleProgress(),
                           progressbar.Bar(marker=["#","█"][six.PY3], left=" |", right="| "),
                           progressbar.ETA()]
                )
            pbar.start()
        else:
            pbar = None

        if jobs > 1:
            pool = multiprocessing.pool.ThreadPool(jobs)
            metalist = pool.map(_parse_file, [(mzml_file, ontology, parser, pbar) for mzml_file in sorted(mzml_files)])
        else:
            metalist = [_parse_file([mzml_file, ontology, parser, pbar]) for mzml_file in sorted(mzml_files)]

        # update isa-tab file
        if merge and extension=='imzML':
            if verbose:
                print('Attempting to merge profile and centroid scans')
            metalist = merge_spectra(metalist)

        if metalist:
            if verbose:
                print("Parsing mzML meta information into ISA-Tab structure")
            isa_tab = ISA_Tab(out_path, study_identifier, usermeta=meta_loader.usermeta, template_directory=template_directory)
            isa_tab.write(metalist, extension, split=split)

    else:
        warnings.warn("No files were found in {}.".format(in_path), UserWarning)

def main(argv=None):
    """Run **mzml2isa** from the command line

    Arguments
        argv (list, optional): the list of arguments to run mzml2isa
            with (if None, then sys.argv is used) [default: None]
    """
    p = argparse.ArgumentParser(prog=__name__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description='''Extract meta information from (i)mzML files and create ISA-tab structure''',
        usage='mzml2isa -i IN_PATH -o OUT_PATH -s STUDY_ID [options]',
    )

    p.add_argument('-i', dest='in_path', help='input folder or archive containing mzML files', required=True)
    p.add_argument('-o', dest='out_path', help='out folder (a new directory will be created here)', required=True)
    p.add_argument('-s', dest='study_id', help='study identifier (e.g. MTBLSxxx)', required=True)
    p.add_argument('-m', dest='usermeta', help='additional user provided metadata (JSON format)', default=None, required=False)#, type=json.loads)
    p.add_argument('-j', dest='jobs', help='launch different processes for parsing', action='store', required=False, default=1, type=int)
    p.add_argument('-n', dest='split', help='do NOT split assay files based on polarity', action='store_false', default=True)
    p.add_argument('-c', dest='merge', help='do NOT group centroid & profile samples', action='store_false', default=True)
    p.add_argument('-W', dest='wrng_ctrl', help='warning control (with python default behaviour)', action='store', default='once', required=False, choices=['ignore', 'always', 'error', 'default', 'module', 'once'])
    p.add_argument('-t', dest='template_dir', help='directory containing default template files', action='store', default=None)
    p.add_argument('--version', action='version', version='mzml2isa {}'.format(__version__))
    p.add_argument('-v', dest='verbose', help="show more output (default if progressbar2 is not installed)", action='store_true', default=False)

    args = p.parse_args(argv or sys.argv[1:])


    if not progressbar:
        setattr(args, 'verbose', True)

    if args.verbose:
        print("{} input path: {}".format(os.linesep, args.in_path))
        print("output path: {}".format(os.path.join(args.out_path, args.study_id)))
        print("Sample identifier:{}{}".format(args.study_id, os.linesep))

    with warnings.catch_warnings():
        warnings.filterwarnings(args.wrng_ctrl)
        convert(args.in_path, args.out_path, args.study_id,
           usermeta=args.usermeta, split=args.split,
           merge=args.merge, verbose=args.verbose,
           jobs=args.jobs, template_directory=args.template_dir
        )



#### DEPRECATED

@functools.wraps(main)
def run(*args, **kwargs):
    warnings.warn("mzml2isa.parsing.run is deprecated, use "
                  "mzml2isa.parsing.main instead", DeprecationWarning)
    main(*args, **kwargs)

@functools.wraps(convert)
def full_parse(*args, **kwargs):
    warnings.warn("mzml2isa.parsing.full_parse is deprecated, use "
                  "mzml2isa.parsing.convert instead", DeprecationWarning)
    full_parse(*args, **kwargs)


if __name__ == '__main__':
    main()





"""
Content
-----------------------------------------------------------------------------
This module contains a single class, ISA_Tab, which is used to dump a list
of mzML.meta or imzML.meta dictionaries to ISA files.

About
-----------------------------------------------------------------------------
The mzml2isa parser was created by Tom Lawson (University of Birmingham, UK)
as part of a NERC funded placement at EBI Cambridge in June 2015. Python 3
port and enhancements were carried out by Martin Larralde (ENS Cachan, FR)
in June 2016 during an internship at the EBI Cambridge.

License
-----------------------------------------------------------------------------
GNU General Public License version 3.0 (GPLv3)
"""
from __future__ import absolute_import

import os
import csv
import sys
import functools

from . import (
    __author__,
    __name__,
    __version__,
    __license__
)
from .utils import (
    PermissiveFormatter,
    TEMPLATES_DIR,
    _ChainMap
)


class ISA_Tab(object):
    """Class to export a list of mzML or imzML metadata dictionnaries to ISA-Tab files

    Attributes:
        usermeta (dict): a dictionary containing metadata defined
            by the user, such as "Study Publication" or "Submission Date". Defaults
            to None.
        isa_env (dict): a dictionary containing various environment
            variables associated with the current ISA_Tab object (such as
            'Study file name', 'Written assays', etc.)
    """

    def __init__(self, out_dir, name, **kwargs):
        """Setup the environments and the directories

        Arguments:
            out_dir (str): the path to the output directory
            name (str): the name of the *omics study to generate

        Keyword Arguments:
            usermeta (dict, optional): a dictionary containing metadata defined
                by the user, such as "Study Publication" or "Submission Date".
                [default: None]
            template_directory (str, optional): the path to a directory containing
                custom ISA-Tab templates. No all templates are required, so if for instance
                only your "a_imzML.txt" is non-standard, then you only have to have a new
                "a_imzML.txt" in your custom template directory. If None, the uses the
                ones shipping with mzml2isa, compatible with MetaboLights [default: None]
        """
        usermeta = kwargs.get('usermeta', None)
        template_directory = kwargs.get('template_directory', None)

        # Create one or several study files / one or several study section in investigation
        self.usermeta = usermeta or {}
        self.isa_env = {
            'out_dir': os.path.join(out_dir, name),
            'Study Identifier':  name,
            'Study file name': 's_{}.txt'.format(name),
            'Assay polar file name': 'a_{}_{{}}_metabolite_profiling_mass_spectrometry.txt'.format(name),
            'Assay file name': 'a_{}_metabolite_profiling_mass_spectrometry.txt'.format(name),
            'default_path': TEMPLATES_DIR,
            'template_path': template_directory or TEMPLATES_DIR,
            'Technology type': [],
            'Measurement type': [],
            'Written assays': [],
            'mzML measurement': {'name': 'metabolite profiling', 'accession':'http://purl.obolibrary.org/obo/OBI_0000366', 'ref':'OBI' },
            'mzML technology': {'name': 'mass spectrometry', 'accession':'http://purl.obolibrary.org/obo/OBI_0000470', 'ref':'OBI' },

        }

        self.isa_env['Converter'] = __name__
        self.isa_env['Converter version'] = __version__

    def write(self, metalist, datatype, **kwargs):
        """Generate and write the ISA files

        Arguments:
            metalist (list): a list of mzml or imzml metadata dictionaries
            datatype (str): the datatype of the study (either 'mzML' or 'imzML')

        Keyword Arguments:
            split (bool, optional): a boolean stating if assay files should be split
                based on their polarities. [default: True]
        """
        split=kwargs.get('split', True)

        self.isa_env['Platform'] = [ next((meta['Instrument'] for meta in metalist if 'Instrument' in meta), '') ]

        if not os.path.exists(self.isa_env['out_dir']):
            os.makedirs(self.isa_env['out_dir'])

        h,d = self.make_assay_template(metalist, datatype)

        self.create_assay(metalist, h, d, split=split)
        self.create_study(metalist,datatype)
        self.create_investigation(metalist, datatype)

    def make_assay_template(self, metalist, datatype):
        """Generate the assay template rows

        Parameters:
            metalist (list): a list of mzml or imzml metadata dictionaries
            datatype (str): the datatype of the study (either 'mzML' or 'imzML')

        Returns
            tuple: a tuple containg the list containing the assay headers row
                and the list containing the assay data row based on the cardinality
                of elements in the metalist
        """

        template_a_path = os.path.join(self.isa_env['template_path'], 'a_{}.txt'.format(datatype))
        if not os.path.exists(template_a_path):
            template_a_path = os.path.join(self.isa_env['template_path'], 'a_{}.txt'.format(datatype))

        with open(template_a_path, 'r') as a_in:
            headers, data = [x.strip().replace('"', '').split('\t') for x in a_in.readlines()]

        i = 0
        while i < len(headers):
            header, datum = headers[i], data[i]

            if '{{' in datum and 'Term' not in header:
                entry_list = metalist[0][self.unparameter(header)]['entry_list']
                hsec, dsec = (headers[i:i+3], data[i:i+3]) \
                                if headers[i+1] == "Term Source REF" \
                                else (headers[i:i+1], data[i:i+1])

                headers[:] = headers[:i] + headers[i+len(hsec):] # Remove the sections we are
                data[:] =    data[:i]    +    data[i+len(dsec):] # going to format and insert

                for n in reversed(range(len(entry_list))):
                    for (h,d) in zip(reversed(hsec),reversed(dsec)):
                        headers.insert(i, h)
                        data.insert(i, d.format(n))

            i+= 1

        return headers, data

    def create_assay(self, metalist, headers, data, **kwargs):
        """Write the assay file

        Arguments:
            metalist (list): a list of mzml or imzml metadata dictionaries
            datatype (str): the datatype of the study (either 'mzML' or 'imzML')
            headers (list): the list containing the assay headers row

        Keyword Arguments:
            split (bool, optional): a boolean stating if assay files should be split
                based on their polarities. [default: True]
        """
        split = kwargs.get('split', True)
        fmt = PermissiveFormatter()

        if split:
            polarities = set( meta['Scan polarity']['name'] for meta in metalist ) \
                            if 'Scan polarity' in metalist[0] else ['nopolarity']
        else:
            polarities = ['nosplit']

        new_a_path = os.path.join(self.isa_env['out_dir'], self.isa_env['Assay file name']) \
                        if len(polarities)==1 \
                        else os.path.join(self.isa_env['out_dir'], self.isa_env['Assay polar file name'])

        for polarity in polarities:

            csv_wopen = functools.partial(open, mode='w', newline='') \
                        if sys.version_info[0]==3 \
                        else functools.partial(open, mode='wb')

            with csv_wopen(new_a_path.format(polarity[:3].upper())) as a_out:

                self.isa_env['Written assays'].append(os.path.basename(new_a_path.format(polarity[:3].upper())))
                self.isa_env['Technology type'].append(self.isa_env['mzML technology'])
                self.isa_env['Measurement type'].append(self.isa_env['mzML measurement'])

                writer=csv.writer(a_out, quotechar='"', quoting=csv.QUOTE_ALL, delimiter='\t')
                writer.writerow(headers)

                for meta in ( x for x in metalist if x['Scan polarity']['name']==polarity ):
                    writer.writerow( [ fmt.vformat(x, None, _ChainMap(meta, self.usermeta)) for x in data] )

    def create_study(self, metalist, datatype):
        """Write the study file

        Arguments:
            metalist (:obj:`list`): a list of mzml or imzml metadata dictionaries
            datatype (:obj:`str`): the datatype of the study (either 'mzML' or 'imzML')
        """
        template_s_path = os.path.join(self.isa_env['template_path'], 's_{}.txt'.format(datatype))
        if not os.path.exists(template_s_path):
            template_s_path = os.path.join(self.isa_env['default_path'], 's_{}.txt'.format(datatype))

        new_s_path = os.path.join(self.isa_env['out_dir'], self.isa_env['Study file name'])

        fmt = PermissiveFormatter()

        with open(template_s_path, 'r') as s_in:
            headers, data = s_in.readlines()

        with open(new_s_path, 'w') as s_out:
            s_out.write(headers)
            for meta in metalist:
                s_out.write(fmt.vformat(data, None, _ChainMap(meta, self.usermeta)))

    def create_investigation(self, metalist, datatype):
        """Write the investigation file

        Arguments:
            metalist (:obj:`list`): a list of mzml or imzml metadata dictionaries
            datatype (:obj:`str`): the datatype of the study (either 'mzML' or 'imzML')
        """
        template_i_path = os.path.join(self.isa_env['template_path'], 'i_{}.txt'.format(datatype))
        if not os.path.exists(template_i_path):
            template_i_path = os.path.join(self.isa_env['default_path'], 'i_{}.txt'.format(datatype))

        new_i_path = os.path.join(self.isa_env['out_dir'], 'i_Investigation.txt')

        meta = metalist[0]
        fmt = PermissiveFormatter()

        chained = _ChainMap(self.isa_env, meta, self.usermeta)

        with open(template_i_path, 'r') as i_in:
            with open(new_i_path, "w") as i_out:
                for l in i_in:

                    if "{{" in l:
                        l, value = l.strip().split('\t')
                        label = value[3:].split('[')[0]

                        if label in chained:

                            for k in range(len(chained[label])):
                                l = '\t'.join([l, value.format(k)])
                            l += '\n'
                        else:
                            l = "\t".join([l, '\"\"', '\n'])

                    l = fmt.vformat(l, None, chained)
                    i_out.write(l)

    @staticmethod
    def unparameter(string):
        """Extract string 's' from 'Parameter Value[s]'

        Arguments:
            string (str): full string

        Return:
            str: the extracted substring
        """
        return string.replace('Parameter Value[', '').replace(']', '')



"""
Content
-----------------------------------------------------------------------------
This module includes some tricks for switching the behaviour of the script
depending on the Python version, while keeping a proper API in the core of
the other modules, as well as miscellaneous class and function definitions.

About
-----------------------------------------------------------------------------
The mzml2isa parser was created by Tom Lawson (University of Birmingham, UK)
as part of a NERC funded placement at EBI Cambridge in June 2015. Python 3
port and enhancements were carried out by Martin Larralde (ENS Cachan, FR)
in June 2016 during an internship at the EBI Cambridge.

License
-----------------------------------------------------------------------------
GNU General Public License version 3.0 (GPLv3)
"""

import collections
import os
import sys
import six
import copy
import zipfile
import tarfile
import functools
import string
import warnings
import pronto
import itertools

from . import (
    __author__,
    __name__,
    __version__,
    __license__,
)

## RESOURCES
TEMPLATES_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates")
ONTOLOGIES_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "ontologies")
MS_CV_URL = 'https://raw.githubusercontent.com/HUPO-PSI/psi-ms-CV/master/psi-ms.obo'
IMS_CV_URL = 'https://raw.githubusercontent.com/ISA-tools/mzml2isa/master/mzml2isa/imagingMS.obo'

## AVAILABLE XML PARSER
try:
    from lxml import etree

    def get_parent(element, tree):
        """Finds every parent of a tree node.

        Uses the method provided by lxml.etree
        """
        return element.getparent()

except ImportError:

    try:
        from xml.etree import cElementTree as etree
    except ImportError:
        from xml.etree import ElementTree as etree

    def get_parent(element, tree):
        """Finds every parent of a tree node.

        As xml.ElementTree has no **.getparent** method, the following was
        proposed here : http://stackoverflow.com/questions/2170610#20132342
        """
        # {c:p for p in tree.iter() for c in p}[element]
        # next(p for p in tree.iter() for c in p if c==element)
        return next(p for p in tree.iter() if element in p)


## VERSION AGNOSTIC UTILS
class PermissiveFormatter(string.Formatter):
    """A formatter that replace wrong and missing key with a blank."""
    def __init__(self, missing='', bad_fmt=''):
        self.missing, self.bad_fmt=missing, bad_fmt

    def get_field(self, field_name, args, kwargs):
        # Handle a key not found
        try:
            val=super(PermissiveFormatter, self).get_field(field_name, args, kwargs)
        except (KeyError, AttributeError, IndexError, TypeError):
            val=None,field_name
        return val

    def format_field(self, value, spec):
        # handle an invalid format
        if value==None:
            return self.missing
        try:
            return super(PermissiveFormatter, self).format_field(value, spec)
        except ValueError:
            if self.bad_fmt is not None:
                return self.bad_fmt
            else:
                raise

class _TarFile(tarfile.TarFile):
    """A TarFile proxy with a setable name
    """

    def __init__(self, name, buffered_reader):
        self.name = name
        self.BufferedReader = buffered_reader

    def __getattr__(self, attr):
        if attr=="name":
            return self.name
        return getattr(self.BufferedReader, attr)

class _ChainMap(collections.Mapping):
    """A quick backport of collections.ChainMap
    """

    def __init__(self, *maps):
        self.maps = list(maps)

    def __getitem__(self, key):
        for mapping in self.maps:
            try:
                return mapping[key]
            except KeyError:
                pass
        return self.__missing__(key)

    @staticmethod
    def __missing__(key):
        raise KeyError(key)

    def __iter__(self):
        return itertools.chain(*self.mappings)

    def __len__(self):
        return sum(len(x) for x in self.mappings)

def merge_spectra(metalist):
    """Merge centroid and spectrum metadata of a same sample
    """
    profiles = [m for m in metalist \
        if m['Spectrum representation']['entry_list'][0]['name']=='profile spectrum']
    centroid = [m for m in metalist \
        if m['Spectrum representation']['entry_list'][0]['name']=='centroid spectrum']

    profiles.sort(key=lambda x: x['Sample Name']['value'])
    centroid.sort(key=lambda x: x['Sample Name']['value'])

    if len(profiles)!=len(centroid):
        return metalist

    for p,c in zip(profiles, centroid):
        p['Derived Spectral Data File']['entry_list'].extend(
            c['Derived Spectral Data File']['entry_list']
        )
        p['Raw Spectral Data File']['entry_list'].extend(
            c['Raw Spectral Data File']['entry_list']
        )
        p['Spectrum representation']['entry_list'].extend(
            c['Spectrum representation']['entry_list']
        )
        p['Sample Name']['value'] = longest_substring(
            p['Sample Name']['value'],c['Sample Name']['value']
        ).strip('-_;:() \n\t')
        p['MS Assay Name']['value'] = p['Sample Name']['value']

    return profiles

def dict_update(d, u):
    """Update a nested dictionnary of various depth

    Shamelessly taken from here: http://stackoverflow.com/a/3233356/623424
    And updated to work with dictionaries nested in lists.
    """
    for k, v in six.iteritems(u):
        if isinstance(v, collections.Mapping):
            if not k in d:
                warnings.warn("Unrecognized key: {}".format(k), UserWarning)
            r = dict_update(d.get(k, {}), v)
            d[k] = r
        elif isinstance(v, list):
            r = []
            for x in v:                      # v Mandatory because of Python linking lists
                r.append(dict_update(copy.deepcopy(d[k][0]), copy.deepcopy(x)))
            d[k] = r
        else:
            d[k] = u[k]
    return d

def pyxpath(meta, query):
        """Finds every occurence of *query* in *mzMLmeta.tree* with proper namespace

        This function also formats the xpath query using the *mzMLmeta.env
        dictionnary created by the **mzml.mzMLmeta.build_env** function.
        """
        return meta.tree.iterfind(query.format(**meta.env), meta.ns)

def longest_substring(string1, string2):
    answer = ""
    len1, len2 = len(string1), len(string2)
    for i in range(len1):
        match = ""
        for j in range(len2):
            if (i + j < len1 and string1[i + j] == string2[j]):
                match += string2[j]
            else:
                if (len(match) > len(answer)): answer = match
                match = ""
    return answer

def compr_extract(compr_pth):
    """Extract tar.gz or .zip files into Python objects

    Arguments:
        compr_path (str): the path to the compressed file

    Returns:
        tarfile.TarFile: if the file is a gzipped tar file
        zipfile.ZipFile: if the file is a zipped file
    """

    filend = ('.mzml', '.imzml')
    if zipfile.is_zipfile(compr_pth):
        comp = zipfile.ZipFile(compr_pth)
        cfiles = [comp.open(f) for f in comp.namelist() if f.lower().endswith(filend)]
        filelist = [f.filename for f in comp.filelist]
    else:
        comp = tarfile.open(compr_pth, 'r:*')
        cfiles = [_TarFile(m.name, comp.extractfile(m)) for m in comp.getmembers() if m.name.lower().endswith(filend)]
        filelist = [f for f in comp.getnames()]

    # And add these file names as additional attribute the compression tar or zip objects
    for cf in cfiles:
        cf.filelist = filelist

    return cfiles

def star_args(func):
    """Unpack arguments if they come packed
    """
    @functools.wraps(func)
    def new_func(*args):
        if len(args)==1:
            return func(*args[0])
        else:
            return func(*args)
    return new_func

def get_ontology(name):
    """Imports the requested ontology with pronto

    Tries to reach the online version, and if it fails then
    use the local version instead.

    Arguments:
        name (str): the name of the ontology to import (either
            'MS' or 'IMS')
    """
    warnings.simplefilter('ignore', pronto.utils.ProntoWarning)
    if name == 'MS':
        try:
            obo = pronto.Ontology(MS_CV_URL, False)
        except BaseException as be:
            obo = pronto.Ontology(os.path.join(ONTOLOGIES_DIR,"psi-ms.obo"), False)
            warnings.warn("Could not use latest online MS ontology, "
                          "using local (version {})".format(obo.meta['version']))
    elif name == 'IMS':
        try:
            obo = pronto.Ontology(IMS_CV_URL, True, 1)
        except BaseException as be:
            obo = pronto.Ontology(os.path.join(ONTOLOGIES_DIR,"imagingMS.obo"), True, 1)
            warnings.warn("Could not use latest online IMS ontology, "
                          "using local (version {})".format(obo.meta['version']))
    else:
        raise ValueError("Unknow ontology to import: {}".format(name))

    return obo





from __future__ import absolute_import

import json
import openpyxl
import six
import os
import collections

from . import (
    __author__,
    __license__,
    __name__,
    __version__,
)

class UserMetaLoader(object):

    CATEGORIZED_MAP = collections.OrderedDict([

        ("Assay Parameters", collections.OrderedDict([

            ("Chromatography Instrument Name", [
                ("Chromatography Instrument", "name"), False,
            ]),

            ("Chromatography Instrument Accession Number", [
                ("Chromatography", "accession"), False,
            ]),

            ("Chromatography Instrument Term Source REF", [
                ("Chromatography Instrument", "ref"), False,
            ]),

            ("Column model", [
                ("Column model", "value"), False,
            ]),

            ("Column type", [
                ("Column type", "value"), False,
            ]),

            ("Derivatization", [
                ("Derivatization", "value"), False,
            ]),

            ("Post Extraction", [
                ("Post Extraction", "value"), False,
            ]),

        ])),

        ("Characteristics", collections.OrderedDict([

            ("Organism Name", [
                ("characteristics", "organism", "name"), False,
            ]),

            ("Organism Accession Number", [
                ("characteristics", "organism", "accession"), False,
            ]),

            ("Organism Term Source REF", [
                ("characteristics", "organism", "ref"), False,
            ]),

            ("Organism Part Name", [
                ("characteristics", "organism_part", "name"), False,
            ]),

            ("Organism Part Accession Number", [
                ("characteristics", "organism_part", "accession"), False,
            ]),

            ("Organism Part Term Source REF", [
                ("characteristics", "organism_part", "ref"), False,
            ]),

            ("Organism Variant Name", [
                ("characteristics", "organism_variant", "name"), False,
            ]),

            ("Organism Variant Accession Number", [
                ("characteristics", "organism_variant", "accession"), False,
            ]),

            ("Organism Variant Term Source REF", [
                ("characteristics", "organism_variant", "ref"), False,
            ]),

        ])),

        ("Protocol Description", {

            "Chromatography Description": [
                ("description", "chroma"), False,
            ],

            "Data Transformation Description": [
                ("description", "data_trans"), False,
            ],

            "Extraction Description": [
                ("description", "extraction"), False,
            ],

            "Mass Spectrometry Description": [
                ("description", "mass_spec"), False,
            ],

            "Metabolite Identification Description": [
                ("description", "metabo_id"), False,
            ],

            "Sample Collection Description": [
                ("description", "sample_collect"), False,
            ],

            "Investigation Description": [
                ("investigation", "description"), False,
            ],

        }),

        ("Investigation", collections.OrderedDict([

            ("Investigation Identifier", [
                ("investigation", "identifier"), False,
            ]),

            ("Investigation Release Date", [
                ("investigation", "release_date"), False,
            ]),

            ("Investigation Submission Date", [
                ("investigation", "submission_date"), False,
            ]),

            ("Investigation Publication Authors", [
                ("investigation_publication", "author_list"), False,
            ]),

            ("Investigation Publication Title", [
                ("investigation_publication", "title"), False,
            ]),

            ("Investigation Publication DOI", [
                ("investigation_publication", "doi"), False,
            ]),

            ("Investigation Publication Pubmed ID", [
                ("investigation_publication", "pubmed"), False,
            ]),

            ("Investigation Publication Status Name", [
                ("investigation_publication", "status", "name"), False,
            ]),

            ("Investigation Publication Status Accession Number", [
                ("investigation_publication", "status", "accession"), False,
            ]),

            ("Investigation Publication Status Term Source REF", [
                ("investigation_publication", "status", "ref"), False,
            ])

        ])),

        ("Study", collections.OrderedDict([

            ("Study Description", [
                ("study", "description"), False,
            ]),

            ("Study Indentifier", [
                ("study", "identifier"), False,
            ]),

            ("Study Release Date", [
                ("study", "release_date"), False,
            ]),

            ("Study Submission Date", [
                ("study", "submission_date"), False,
            ]),

            ("Study Title", [
                ("study", "title"), False,
            ]),

            ("Study Publication Authors", [
                ("study_publication", "author_list"), False,
            ]),

            ("Study Publication Title", [
                ("study_publication", "title"), False,
            ]),

            ("Study Publication DOI", [
                ("study_publication", "doi"), False,
            ]),

            ("Study Publication Pubmed ID", [
                ("study_publication", "pubmed"), False,
            ]),

            ("Study Publication Status Name", [
                ("study_publication", "status", "name"), False,
            ]),

            ("Study Publication Status Accession Number", [
                ("study_publication", "status", "accession"), False,
            ]),

            ("Study Publication Status Term Source REF", [
                ("study_publication", "status", "ref"), False,
            ]),

        ])),

        ("Contacts", collections.OrderedDict([

            ("Investigation Contacts First Name", [
                ("investigation_contacts", "first_name"), True
            ]),

            ("Investigation Contacts Middle Name", [
                ("investigation_contacts", "mid"), True,
            ]),

            ("Investigation Contacts Last Name", [
                ("investigation_contacts", "last_name"), True,
            ]),

            ("Investigation Contacts Affiliation", [
                ("investigation_contacts", "affiliation"), True,
            ]),

            ("Investigation Contacts Adress", [
                ("investigation_contacts", "adress"), True,
            ]),

            ("Investigation Contacts Email", [
                ("investigation_contacts", "email"), True,
            ]),

            ("Investigation Contacts Phone", [
                ("investigation_contacts", "phone"), True,
            ]),

            ("Investigation Contacts Fax", [
                ("investigation_contacts", "fax"), True,
            ]),

            ("Investigation Contacts Role Name", [
                ("investigation_contacts", "roles", "name"), True,
            ]),

            ("Investigation Contacts Role Term Source REF", [
                ("investigation_contacts", "roles", "ref"), True,
            ]),

            ("Investigation Contacts Role Accession Number", [
                ("investigation_contacts", "roles", "accession"), True,
            ]),

            ("Study Contacts First Name", [
                ("study_contacts", "first_name"), True
            ]),

            ("Study Contacts Middle Name", [
                ("study_contacts", "mid"), True,
            ]),

            ("Study Contacts Last Name", [
                ("study_contacts", "last_name"), True,
            ]),

            ("Study Contacts Affiliation", [
                ("study_contacts", "affiliation"), True,
            ]),

            ("Study Contacts Adress", [
                ("study_contacts", "adress"), True,
            ]),

            ("Study Contacts Email", [
                ("study_contacts", "email"), True,
            ]),

            ("Study Contacts Phone", [
                ("study_contacts", "phone"), True,
            ]),

            ("Study Contacts Fax", [
                ("study_contacts", "fax"), True,
            ]),

            ("Study Contacts Role Name", [
                ("study_contacts", "roles", "name"), True,
            ]),

            ("Study Contacts Role Term Source REF", [
                ("study_contacts", "roles", "ref"), True,
            ]),

            ("Study Contacts Role Accession Number", [
                ("study_contacts", "roles", "accession"), True,
            ]),

        ]))
    ])

    MAP = {k:v
            for submap in six.itervalues(CATEGORIZED_MAP)
                for k,v in six.iteritems(submap)
                }

    def __init__(self, usermeta_token):


        if usermeta_token is None:
            self.usermeta = None

        elif usermeta_token.endswith('.xlsx'):

            self._parse_xlsx_file(usermeta_token)


        elif usermeta_token.endswith('.json'):
            self._parse_json_file(usermeta_token)

        else:
            self._parse_json_stdin(usermeta_token)

    def _parse_json_file(self, usermeta_token):
        """
        """
        try:
            with open(args.usermeta) as f:
                self.usermeta = json.load(f)
        except json.decoder.JSONDecodeError:
            self.usermeta = None
            warnings.warn("JSON usermeta could not be parsed from {}.".format(usermeta_token))
        except OSError:
            self.usermeta = None
            warnings.warn("File {} not found.".format(usermeta_token))

    def _parse_json_stdin(self, usermeta_token):
        """
        """
        try:
            usermeta = json.loads(args.usermeta)
        except json.decoder.JSONDecodeError:
            self.usermeta = None
            warnings.warn("JSON usermeta could not be parsed from <stdin>.")

    def _parse_xlsx_file(self, usermeta_token):
        """
        """
        self.usermeta = {}
        sheet = openpyxl.load_workbook(usermeta_token).worksheets[0]

        for row in sheet.iter_rows():

            # if the row is [Header, Value, None, ..., None] -> non multiple values
            if row[1] != None and all(x.value is None for x in row[2:]):
                header = row[0].value
                value = row[1].value
            # if the row is [Header, Value, None ... None, Value, None ... None, ...] -> multiple values
            else:
                header = row[0].value
                value = [x.value if x.value is not None else '' for x in row[1:]]

            # Skip line if comment or empty headers
            if header is None or header.startswith("#") or not value:
                continue

            # Check in map how to translate excel headers to the metatadata dict
            true_name, more_than_one = self.MAP[header]

            # if there's only one value to write: find the dict to update
            # (self.usermeta[key1][key2][...][keyn])
            if not more_than_one:
                item_to_set = self.usermeta
                for i,path_node in enumerate(true_name[:-1]):
                    item_to_set = item_to_set.setdefault(path_node, {true_name[i+1]: {}})
                if isinstance(value, list):
                    item_to_set[true_name[-1]] = ", ".join(value or [])
                else:
                    item_to_set[true_name[-1]] = value or ''

            # if there's only one value to write: find the dict to update
            # (self.usermeta[key1][i][key2][...][keyn] where i is the offset
            # of the current value)
            else:
                for i, value in enumerate(value):
                    item_to_set = self.usermeta.setdefault(true_name[0], [])
                    if len(item_to_set) <= i:
                        item_to_set.append({})
                    item_to_set = self.usermeta[true_name[0]][i]

                    for i,path_node in enumerate(true_name[1:-1]):
                        item_to_set = item_to_set.setdefault(path_node, {true_name[i+1]: {}})

                    item_to_set[true_name[-1]] = value

        # Remove empty multiple_values dictionaries
        for mv_key in (v[0][0] for v in six.itervalues(self.MAP) if v[1]):
            try:
                for value in self.usermeta[mv_key]:
                    empty = not any(v for k,v in six.iteritems(value)
                                        if not isinstance(v, collections.Mapping))
                    if empty:
                        self.usermeta[mv_key].remove(value)
            except KeyError:
                pass

    @classmethod
    def dump_template_xlsx(cls, output_directory, name='usermeta.xlsx'):

        wb = openpyxl.Workbook()
        x, y = 65,1

        for category, submap in six.iteritems(cls.CATEGORIZED_MAP):
            y += 1
            wb.worksheets[0]['{}{}'.format(chr(x), y)] = "#### {} ####".format(category.upper())
            y += 1
            for header in submap:
                wb.worksheets[0]['{}{}'.format(chr(x), y)] = header
                y += 1

        wb.save(os.path.join(output_directory, name))


if __name__=="__main__":
    UserMetaLoader.dump_template_xlsx(os.getcwd())
